"""Utility functions for pre-commit hooks."""

import itertools
import os
from pathlib import Path
from typing import Sequence


def find_project_root(srcs: Sequence[str]):
    """
    Return a directory containing .git, .hg, pyproject.toml, or setup.cfg.

    That directory can be one of the directories passed in ``srcs`` or their
    common parent. If no directory in the tree contains a marker that would
    specify it's the project root, the root of the file system is returned.

    Parameters
    ----------
    srcs : Sequence[str]
        The filepaths to run the hook on.

    Returns
    -------
    str
        The project root directory.

    See Also
    --------
    black.find_project_root :
        This function was adapted from
        `Black <https://github.com/psf/black/blob/main/src/black/files.py>`_.
    """
    if not srcs:
        return Path(".").resolve(), "current directory"

    common_path = Path(
        os.path.commonpath([Path(src).expanduser().resolve() for src in srcs])
    )

    for dir in itertools.chain([common_path], common_path.parents):
        if (dir / "pyproject.toml").is_file():
            return dir, "pyproject.toml"
        if (dir / "setup.cfg").is_file():
            return dir, "setup.cfg"
        if (dir / ".git").exists() or (dir / ".hg").is_dir():
            return dir, "version control"

    return dir, "file system root"



from collections import namedtuple
from copy import deepcopy
import re
import textwrap
import warnings

import jinja2

from numpydoc.numpydoc import update_config
from numpydoc.xref import DEFAULT_LINKS
from numpydoc.docscrape import NumpyDocString, FunctionDoc, ClassDoc, ParseError
from numpydoc.docscrape_sphinx import (
    SphinxDocString,
    SphinxClassDoc,
    SphinxFunctionDoc,
    get_doc_object,
)
import pytest
from pytest import raises as assert_raises
from pytest import warns as assert_warns


doc_txt = """\
  numpy.multivariate_normal(mean, cov, shape=None, spam=None)

  Draw values from a multivariate normal distribution with specified
  mean and covariance.

  The multivariate normal or Gaussian distribution is a generalisation
  of the one-dimensional normal distribution to higher dimensions.

  Parameters
  ----------
  mean : (N,) ndarray
      Mean of the N-dimensional distribution.

      .. math::

         (1+2+3)/3

  cov : (N, N) ndarray
      Covariance matrix of the distribution.
  shape : tuple of ints
      Given a shape of, for example, (m,n,k), m*n*k samples are
      generated, and packed in an m-by-n-by-k arrangement.  Because
      each sample is N-dimensional, the output shape is (m,n,k,N).
  dtype : data type object, optional (default : float)
      The type and size of the data to be returned.

  Returns
  -------
  out : ndarray
      The drawn samples, arranged according to `shape`.  If the
      shape given is (m,n,...), then the shape of `out` is
      (m,n,...,N).

      In other words, each entry ``out[i,j,...,:]`` is an N-dimensional
      value drawn from the distribution.
  list of str
      This is not a real return value.  It exists to test
      anonymous return values.
  no_description

  Other Parameters
  ----------------
  spam : parrot
      A parrot off its mortal coil.

  Raises
  ------
  RuntimeError
      Some error

  Warns
  -----
  RuntimeWarning
      Some warning

  Warnings
  --------
  Certain warnings apply.

  Notes
  -----
  Instead of specifying the full covariance matrix, popular
  approximations include:

    - Spherical covariance (`cov` is a multiple of the identity matrix)
    - Diagonal covariance (`cov` has non-negative elements only on the diagonal)

  This geometrical property can be seen in two dimensions by plotting
  generated data-points:

  >>> mean = [0,0]
  >>> cov = [[1,0],[0,100]] # diagonal covariance, points lie on x or y-axis

  >>> x,y = multivariate_normal(mean,cov,5000).T
  >>> plt.plot(x,y,'x'); plt.axis('equal'); plt.show()

  Note that the covariance matrix must be symmetric and non-negative
  definite.

  References
  ----------
  .. [1] A. Papoulis, "Probability, Random Variables, and Stochastic
         Processes," 3rd ed., McGraw-Hill Companies, 1991
  .. [2] R.O. Duda, P.E. Hart, and D.G. Stork, "Pattern Classification,"
         2nd ed., Wiley, 2001.

  See Also
  --------
  some, other, funcs
  otherfunc : relationship
  :py:meth:`spyder.widgets.mixins.GetHelpMixin.show_object_info`

  Examples
  --------
  >>> mean = (1,2)
  >>> cov = [[1,0],[1,0]]
  >>> x = multivariate_normal(mean,cov,(3,3))
  >>> print(x.shape)
  (3, 3, 2)

  The following is probably true, given that 0.6 is roughly twice the
  standard deviation:

  >>> print(list((x[0, 0, :] - mean) < 0.6))
  [True, True]

  .. index:: random
     :refguide: random;distributions, random;gauss

  """


@pytest.fixture(params=["", "\n    "], ids=["flush", "newline_indented"])
def doc(request):
    return NumpyDocString(request.param + doc_txt)


doc_yields_txt = """
Test generator

Yields
------
a : int
    The number of apples.
b : int
    The number of bananas.
int
    The number of unknowns.
"""
doc_yields = NumpyDocString(doc_yields_txt)


doc_sent_txt = """
Test generator

Yields
------
a : int
    The number of apples.

Receives
--------
b : int
    The number of bananas.
c : int
    The number of oranges.

"""
doc_sent = NumpyDocString(doc_sent_txt)


def test_signature(doc):
    assert doc["Signature"].startswith("numpy.multivariate_normal(")
    assert doc["Signature"].endswith("spam=None)")


def test_summary(doc):
    assert doc["Summary"][0].startswith("Draw values")
    assert doc["Summary"][-1].endswith("covariance.")


def test_extended_summary(doc):
    assert doc["Extended Summary"][0].startswith("The multivariate normal")


def test_parameters(doc):
    assert len(doc["Parameters"]) == 4
    names = [n for n, _, _ in doc["Parameters"]]
    assert all(a == b for a, b in zip(names, ["mean", "cov", "shape"]))

    arg, arg_type, desc = doc["Parameters"][1]
    assert arg_type == "(N, N) ndarray"
    assert desc[0].startswith("Covariance matrix")
    assert doc["Parameters"][0][-1][-1] == "   (1+2+3)/3"

    arg, arg_type, desc = doc["Parameters"][2]
    assert arg == "shape"
    assert arg_type == "tuple of ints"
    assert desc[0].startswith("Given")
    assert doc["Parameters"][0][-1][-1] == "   (1+2+3)/3"

    arg, arg_type, desc = doc["Parameters"][3]
    assert arg == "dtype"
    assert arg_type == "data type object, optional (default : float)"
    assert desc[0].startswith("The type and size")


def test_other_parameters(doc):
    assert len(doc["Other Parameters"]) == 1
    assert [n for n, _, _ in doc["Other Parameters"]] == ["spam"]
    arg, arg_type, desc = doc["Other Parameters"][0]
    assert arg_type == "parrot"
    assert desc[0].startswith("A parrot off its mortal coil")


def test_returns(doc):
    assert len(doc["Returns"]) == 3
    arg, arg_type, desc = doc["Returns"][0]
    assert arg == "out"
    assert arg_type == "ndarray"
    assert desc[0].startswith("The drawn samples")
    assert desc[-1].endswith("distribution.")

    arg, arg_type, desc = doc["Returns"][1]
    assert arg == ""
    assert arg_type == "list of str"
    assert desc[0].startswith("This is not a real")
    assert desc[-1].endswith("anonymous return values.")

    arg, arg_type, desc = doc["Returns"][2]
    assert arg == ""
    assert arg_type == "no_description"
    assert not "".join(desc).strip()


def test_yields():
    section = doc_yields["Yields"]
    assert len(section) == 3
    truth = [
        ("a", "int", "apples."),
        ("b", "int", "bananas."),
        ("", "int", "unknowns."),
    ]
    for (arg, arg_type, desc), (arg_, arg_type_, end) in zip(section, truth):
        assert arg == arg_
        assert arg_type == arg_type_
        assert desc[0].startswith("The number of")
        assert desc[0].endswith(end)


def test_sent():
    section = doc_sent["Receives"]
    assert len(section) == 2
    truth = [("b", "int", "bananas."), ("c", "int", "oranges.")]
    for (arg, arg_type, desc), (arg_, arg_type_, end) in zip(section, truth):
        assert arg == arg_
        assert arg_type == arg_type_
        assert desc[0].startswith("The number of")
        assert desc[0].endswith(end)


def test_returnyield():
    doc_text = """
Test having returns and yields.

Returns
-------
int
    The number of apples.

Yields
------
a : int
    The number of apples.
b : int
    The number of bananas.

"""
    assert_raises(ValueError, NumpyDocString, doc_text)


def test_section_twice():
    doc_text = """
Test having a section Notes twice

Notes
-----
See the next note for more information

Notes
-----
That should break...
"""
    with pytest.raises(ValueError, match="The section Notes appears twice"):
        NumpyDocString(doc_text)

    # if we have a numpydoc object, we know where the error came from
    class Dummy:
        """
        Dummy class.

        Notes
        -----
        First note.

        Notes
        -----
        Second note.

        """

        def spam(self, a, b):
            """Spam\n\nSpam spam."""
            pass

        def ham(self, c, d):
            """Cheese\n\nNo cheese."""
            pass

    def dummy_func(arg):
        """
        Dummy function.

        Notes
        -----
        First note.

        Notes
        -----
        Second note.
        """

    with pytest.raises(ValueError, match="Dummy class"):
        SphinxClassDoc(Dummy)

    with pytest.raises(ValueError, match="dummy_func"):
        SphinxFunctionDoc(dummy_func)


def test_notes(doc):
    assert doc["Notes"][0].startswith("Instead")
    assert doc["Notes"][-1].endswith("definite.")
    assert len(doc["Notes"]) == 17


def test_references(doc):
    assert doc["References"][0].startswith("..")
    assert doc["References"][-1].endswith("2001.")


def test_examples(doc):
    assert doc["Examples"][0].startswith(">>>")
    assert doc["Examples"][-1].endswith("True]")


def test_index(doc):
    assert doc["index"]["default"] == "random"
    assert len(doc["index"]) == 2
    assert len(doc["index"]["refguide"]) == 2


def _strip_blank_lines(s):
    "Remove leading, trailing and multiple blank lines"
    s = re.sub(r"^\s*\n", "", s)
    s = re.sub(r"\n\s*$", "", s)
    s = re.sub(r"\n\s*\n", r"\n\n", s)
    return s


def line_by_line_compare(a, b, n_lines=None):
    a = textwrap.dedent(a)
    b = textwrap.dedent(b)
    a = [l.rstrip() for l in _strip_blank_lines(a).split("\n")][:n_lines]
    b = [l.rstrip() for l in _strip_blank_lines(b).split("\n")][:n_lines]
    assert len(a) == len(b)
    for ii, (aa, bb) in enumerate(zip(a, b)):
        assert aa == bb


def test_str(doc):
    # doc_txt has the order of Notes and See Also sections flipped.
    # This should be handled automatically, and so, one thing this test does
    # is to make sure that See Also precedes Notes in the output.
    line_by_line_compare(
        str(doc),
        """numpy.multivariate_normal(mean, cov, shape=None, spam=None)

Draw values from a multivariate normal distribution with specified
mean and covariance.

The multivariate normal or Gaussian distribution is a generalisation
of the one-dimensional normal distribution to higher dimensions.

Parameters
----------
mean : (N,) ndarray
    Mean of the N-dimensional distribution.

    .. math::

       (1+2+3)/3
cov : (N, N) ndarray
    Covariance matrix of the distribution.
shape : tuple of ints
    Given a shape of, for example, (m,n,k), m*n*k samples are
    generated, and packed in an m-by-n-by-k arrangement.  Because
    each sample is N-dimensional, the output shape is (m,n,k,N).
dtype : data type object, optional (default : float)
    The type and size of the data to be returned.

Returns
-------
out : ndarray
    The drawn samples, arranged according to `shape`.  If the
    shape given is (m,n,...), then the shape of `out` is
    (m,n,...,N).

    In other words, each entry ``out[i,j,...,:]`` is an N-dimensional
    value drawn from the distribution.
list of str
    This is not a real return value.  It exists to test
    anonymous return values.
no_description

Other Parameters
----------------
spam : parrot
    A parrot off its mortal coil.

Raises
------
RuntimeError
    Some error

Warns
-----
RuntimeWarning
    Some warning

Warnings
--------
Certain warnings apply.

See Also
--------

`some`_, `other`_, `funcs`_
    ..
`otherfunc`_
    relationship
:py:meth:`spyder.widgets.mixins.GetHelpMixin.show_object_info`
    ..

Notes
-----
Instead of specifying the full covariance matrix, popular
approximations include:

  - Spherical covariance (`cov` is a multiple of the identity matrix)
  - Diagonal covariance (`cov` has non-negative elements only on the diagonal)

This geometrical property can be seen in two dimensions by plotting
generated data-points:

>>> mean = [0,0]
>>> cov = [[1,0],[0,100]] # diagonal covariance, points lie on x or y-axis

>>> x,y = multivariate_normal(mean,cov,5000).T
>>> plt.plot(x,y,'x'); plt.axis('equal'); plt.show()

Note that the covariance matrix must be symmetric and non-negative
definite.

References
----------
.. [1] A. Papoulis, "Probability, Random Variables, and Stochastic
       Processes," 3rd ed., McGraw-Hill Companies, 1991
.. [2] R.O. Duda, P.E. Hart, and D.G. Stork, "Pattern Classification,"
       2nd ed., Wiley, 2001.

Examples
--------
>>> mean = (1,2)
>>> cov = [[1,0],[1,0]]
>>> x = multivariate_normal(mean,cov,(3,3))
>>> print(x.shape)
(3, 3, 2)

The following is probably true, given that 0.6 is roughly twice the
standard deviation:

>>> print(list((x[0, 0, :] - mean) < 0.6))
[True, True]

.. index:: random
   :refguide: random;distributions, random;gauss""",
    )


def test_yield_str():
    line_by_line_compare(
        str(doc_yields),
        """Test generator

Yields
------
a : int
    The number of apples.
b : int
    The number of bananas.
int
    The number of unknowns.
""",
    )


def test_receives_str():
    line_by_line_compare(
        str(doc_sent),
        """Test generator

Yields
------
a : int
    The number of apples.

Receives
--------
b : int
    The number of bananas.
c : int
    The number of oranges.
""",
    )


def test_no_index_in_str():
    assert "index" not in str(
        NumpyDocString(
            """Test idx

    """
        )
    )

    assert "index" in str(
        NumpyDocString(
            """Test idx

    .. index :: random
    """
        )
    )

    assert "index" in str(
        NumpyDocString(
            """Test idx

    .. index ::
        foo
    """
        )
    )


def test_sphinx_str():
    sphinx_doc = SphinxDocString(doc_txt)
    line_by_line_compare(
        str(sphinx_doc),
        """
.. index:: random
   single: random;distributions, random;gauss

Draw values from a multivariate normal distribution with specified
mean and covariance.

The multivariate normal or Gaussian distribution is a generalisation
of the one-dimensional normal distribution to higher dimensions.

:Parameters:

    **mean** : (N,) ndarray
        Mean of the N-dimensional distribution.

        .. math::

           (1+2+3)/3

    **cov** : (N, N) ndarray
        Covariance matrix of the distribution.

    **shape** : tuple of ints
        Given a shape of, for example, (m,n,k), m*n*k samples are
        generated, and packed in an m-by-n-by-k arrangement.  Because
        each sample is N-dimensional, the output shape is (m,n,k,N).

    **dtype** : data type object, optional (default : float)
        The type and size of the data to be returned.

:Returns:

    **out** : ndarray
        The drawn samples, arranged according to `shape`.  If the
        shape given is (m,n,...), then the shape of `out` is
        (m,n,...,N).

        In other words, each entry ``out[i,j,...,:]`` is an N-dimensional
        value drawn from the distribution.

    list of str
        This is not a real return value.  It exists to test
        anonymous return values.

    no_description
        ..

:Other Parameters:

    **spam** : parrot
        A parrot off its mortal coil.

:Raises:

    RuntimeError
        Some error

:Warns:

    RuntimeWarning
        Some warning

.. warning::

    Certain warnings apply.

.. seealso::

    :obj:`some`, :obj:`other`, :obj:`funcs`
        ..
    :obj:`otherfunc`
        relationship
    :py:meth:`spyder.widgets.mixins.GetHelpMixin.show_object_info`
        ..

.. rubric:: Notes

Instead of specifying the full covariance matrix, popular
approximations include:

  - Spherical covariance (`cov` is a multiple of the identity matrix)
  - Diagonal covariance (`cov` has non-negative elements only on the diagonal)

This geometrical property can be seen in two dimensions by plotting
generated data-points:

>>> mean = [0,0]
>>> cov = [[1,0],[0,100]] # diagonal covariance, points lie on x or y-axis

>>> x,y = multivariate_normal(mean,cov,5000).T
>>> plt.plot(x,y,'x'); plt.axis('equal'); plt.show()

Note that the covariance matrix must be symmetric and non-negative
definite.

.. rubric:: References

.. [1] A. Papoulis, "Probability, Random Variables, and Stochastic
       Processes," 3rd ed., McGraw-Hill Companies, 1991
.. [2] R.O. Duda, P.E. Hart, and D.G. Stork, "Pattern Classification,"
       2nd ed., Wiley, 2001.

.. only:: latex

   [1]_, [2]_

.. rubric:: Examples

>>> mean = (1,2)
>>> cov = [[1,0],[1,0]]
>>> x = multivariate_normal(mean,cov,(3,3))
>>> print(x.shape)
(3, 3, 2)

The following is probably true, given that 0.6 is roughly twice the
standard deviation:

>>> print(list((x[0, 0, :] - mean) < 0.6))
[True, True]
""",
    )


def test_sphinx_yields_str():
    sphinx_doc = SphinxDocString(doc_yields_txt)
    line_by_line_compare(
        str(sphinx_doc),
        """Test generator

:Yields:

    **a** : int
        The number of apples.

    **b** : int
        The number of bananas.

    int
        The number of unknowns.
""",
    )


doc2 = NumpyDocString(
    """
    Returns array of indices of the maximum values of along the given axis.

    Parameters
    ----------
    a : {array_like}
        Array to look in.
    axis : {None, integer}
        If None, the index is into the flattened array, otherwise along
        the specified axis"""
)


def test_parameters_without_extended_description():
    assert len(doc2["Parameters"]) == 2


doc3 = NumpyDocString(
    """
    my_signature(*params, **kwds)

    Return this and that.
    """
)


def test_escape_stars():
    signature = str(doc3).split("\n")[0]
    assert signature == r"my_signature(\*params, \*\*kwds)"

    def my_func(a, b, **kwargs):
        pass

    fdoc = FunctionDoc(func=my_func)
    assert fdoc["Signature"] == ""


doc4 = NumpyDocString(
    """a.conj()

    Return an array with all complex-valued elements conjugated."""
)


def test_empty_extended_summary():
    assert doc4["Extended Summary"] == []


doc5 = NumpyDocString(
    """
    a.something()

    Raises
    ------
    LinAlgException
        If array is singular.

    Warns
    -----
    SomeWarning
        If needed
    """
)


def test_raises():
    assert len(doc5["Raises"]) == 1
    param = doc5["Raises"][0]
    assert param.name == ""
    assert param.type == "LinAlgException"
    assert param.desc == ["If array is singular."]


def test_warns():
    assert len(doc5["Warns"]) == 1
    param = doc5["Warns"][0]
    assert param.name == ""
    assert param.type == "SomeWarning"
    assert param.desc == ["If needed"]


# see numpydoc/numpydoc #281
# we want to correctly parse "See Also" both in docstrings both like
# """foo
# and
# """
# foo
@pytest.mark.parametrize("prefix", ["", "\n    "])
def test_see_also(prefix):
    doc6 = NumpyDocString(
        prefix
        + """z(x,theta)

    See Also
    --------
    func_a, func_b, func_c
    func_d : some equivalent func
    foo.func_e : some other func over
             multiple lines
    func_f, func_g, :meth:`func_h`, func_j,
    func_k
    func_f1, func_g1, :meth:`func_h1`, func_j1
    func_f2, func_g2, :meth:`func_h2`, func_j2 : description of multiple
    :obj:`baz.obj_q`
    :obj:`~baz.obj_r`
    :class:`class_j`: fubar
        foobar
    """
    )

    assert len(doc6["See Also"]) == 10
    for funcs, desc in doc6["See Also"]:
        for func, role in funcs:
            if func in (
                "func_a",
                "func_b",
                "func_c",
                "func_f",
                "func_g",
                "func_h",
                "func_j",
                "func_k",
                "baz.obj_q",
                "func_f1",
                "func_g1",
                "func_h1",
                "func_j1",
                "~baz.obj_r",
            ):
                assert not desc, str([func, desc])
            elif func in ("func_f2", "func_g2", "func_h2", "func_j2"):
                assert desc, str([func, desc])
            else:
                assert desc, str([func, desc])

            if func == "func_h":
                assert role == "meth"
            elif func in ("baz.obj_q", "~baz.obj_r"):
                assert role == "obj"
            elif func == "class_j":
                assert role == "class"
            elif func in ("func_h1", "func_h2"):
                assert role == "meth"
            else:
                assert role is None, str([func, role])

            if func == "func_d":
                assert desc == ["some equivalent func"]
            elif func == "foo.func_e":
                assert desc == ["some other func over", "multiple lines"]
            elif func == "class_j":
                assert desc == ["fubar", "foobar"]
            elif func in ("func_f2", "func_g2", "func_h2", "func_j2"):
                assert desc == ["description of multiple"], str(
                    [desc, ["description of multiple"]]
                )


def test_see_also_parse_error():
    text = """
    z(x,theta)

    See Also
    --------
    :func:`~foo`
    """
    with pytest.raises(ValueError, match="See Also entry ':func:`~foo`'"):
        NumpyDocString(text)


def test_see_also_print():
    class Dummy:
        """
        See Also
        --------
        func_a, func_b
        func_c : some relationship
                 goes here
        func_d
        """

        pass

    s = str(FunctionDoc(Dummy, role="func"))
    assert ":func:`func_a`, :func:`func_b`" in s
    assert "    some relationship" in s
    assert ":func:`func_d`" in s


def test_see_also_trailing_comma_warning():
    warnings.filterwarnings("error")
    with assert_warns(
        Warning,
        match="Unexpected comma or period after function list at index 43 of line .*",
    ):
        NumpyDocString(
            """
            z(x,theta)

            See Also
            --------
            func_f2, func_g2, :meth:`func_h2`, func_j2, : description of multiple
            :class:`class_j`: fubar
                foobar
            """
        )


def test_unknown_section():
    doc_text = """
Test having an unknown section

Mope
----
This should be ignored and warned about
"""

    class BadSection:
        """Class with bad section.

        Nope
        ----
        This class has a nope section.
        """

        pass

    with pytest.warns(UserWarning, match="Unknown section Mope") as record:
        NumpyDocString(doc_text)
    assert len(record) == 1

    # SphinxClassDoc has _obj.__name__ == "BadSection". Test that this is
    # included in the message
    msg_match = "Unknown section Nope in the docstring of BadSection"
    with pytest.warns(UserWarning, match=msg_match) as record:
        SphinxClassDoc(BadSection)
    assert len(record) == 1


doc7 = NumpyDocString(
    """

        Doc starts on second line.

        """
)


def test_empty_first_line():
    assert doc7["Summary"][0].startswith("Doc starts")


doc8 = NumpyDocString(
    """

        Parameters with colon and no types:

        Parameters
        ----------

        data :
            some stuff, technically invalid
        """
)


def test_returns_with_roles_no_names():
    """Make sure colons that are part of sphinx roles are not misinterpreted
    as type separator in returns section. See gh-428."""
    docstring = NumpyDocString(
        """
        Returns
        -------
        str or :class:`NumpyDocString`
        """
    )
    expected = "str or :class:`NumpyDocString`"  # not "str or : class:...
    assert docstring["Returns"][0].type == expected
    assert expected in str(docstring)


def test_trailing_colon():
    assert doc8["Parameters"][0].name == "data"


def test_no_summary():
    str(
        SphinxDocString(
            """
    Parameters
    ----------"""
        )
    )


def test_unicode():
    doc = SphinxDocString(
        """
    öäöäöäöäöåååå

    öäöäöäööäååå

    Parameters
    ----------
    ååå : äää
        ööö

    Returns
    -------
    ååå : ööö
        äää

    """
    )
    assert isinstance(doc["Summary"][0], str)
    assert doc["Summary"][0] == "öäöäöäöäöåååå"


def test_plot_examples():
    cfg = dict(use_plots=True)

    doc = SphinxDocString(
        """
    Examples
    --------
    >>> import matplotlib.pyplot as plt
    >>> plt.plot([1,2,3],[4,5,6])
    >>> plt.show()
    """,
        config=cfg,
    )
    assert "plot::" in str(doc), str(doc)

    doc = SphinxDocString(
        """
    Examples
    --------
    >>> from matplotlib import pyplot as plt
    >>> plt.plot([1,2,3],[4,5,6])
    >>> plt.show()
    """,
        config=cfg,
    )
    assert "plot::" in str(doc), str(doc)

    doc = SphinxDocString(
        """
    Examples
    --------
    .. plot::

       import matplotlib.pyplot as plt
       plt.plot([1,2,3],[4,5,6])
       plt.show()
    """,
        config=cfg,
    )
    assert str(doc).count("plot::") == 1, str(doc)


def test_class_members():
    class Dummy:
        """
        Dummy class.

        """

        def spam(self, a, b):
            """Spam\n\nSpam spam."""
            pass

        def ham(self, c, d):
            """Cheese\n\nNo cheese."""
            pass

        @property
        def spammity(self):
            """Spammity index"""
            return 0.95

        class Ignorable:
            """local class, to be ignored"""

            pass

    for cls in (ClassDoc, SphinxClassDoc):
        doc = cls(Dummy, config=dict(show_class_members=False))
        assert "Methods" not in str(doc), (cls, str(doc))
        assert "spam" not in str(doc), (cls, str(doc))
        assert "ham" not in str(doc), (cls, str(doc))
        assert "spammity" not in str(doc), (cls, str(doc))
        assert "Spammity index" not in str(doc), (cls, str(doc))

        doc = cls(Dummy, config=dict(show_class_members=True))
        assert "Methods" in str(doc), (cls, str(doc))
        assert "spam" in str(doc), (cls, str(doc))
        assert "ham" in str(doc), (cls, str(doc))
        assert "spammity" in str(doc), (cls, str(doc))

        if cls is SphinxClassDoc:
            assert ".. autosummary::" in str(doc), str(doc)
        else:
            assert "Spammity index" in str(doc), str(doc)

    class SubDummy(Dummy):
        """
        Subclass of Dummy class.

        """

        def ham(self, c, d):
            """Cheese\n\nNo cheese.\nOverloaded Dummy.ham"""
            pass

        def bar(self, a, b):
            """Bar\n\nNo bar"""
            pass

    for cls in (ClassDoc, SphinxClassDoc):
        doc = cls(
            SubDummy,
            config=dict(show_class_members=True, show_inherited_class_members=False),
        )
        assert "Methods" in str(doc), (cls, str(doc))
        assert "spam" not in str(doc), (cls, str(doc))
        assert "ham" in str(doc), (cls, str(doc))
        assert "bar" in str(doc), (cls, str(doc))
        assert "spammity" not in str(doc), (cls, str(doc))

        if cls is SphinxClassDoc:
            assert ".. autosummary::" in str(doc), str(doc)
        else:
            assert "Spammity index" not in str(doc), str(doc)

        doc = cls(
            SubDummy,
            config=dict(show_class_members=True, show_inherited_class_members=True),
        )
        assert "Methods" in str(doc), (cls, str(doc))
        assert "spam" in str(doc), (cls, str(doc))
        assert "ham" in str(doc), (cls, str(doc))
        assert "bar" in str(doc), (cls, str(doc))
        assert "spammity" in str(doc), (cls, str(doc))

        if cls is SphinxClassDoc:
            assert ".. autosummary::" in str(doc), str(doc)
        else:
            assert "Spammity index" in str(doc), str(doc)


def test_duplicate_signature():
    # Duplicate function signatures occur e.g. in ufuncs, when the
    # automatic mechanism adds one, and a more detailed comes from the
    # docstring itself.

    doc = NumpyDocString(
        """
    z(x1, x2)

    z(a, theta)
    """
    )

    assert doc["Signature"].strip() == "z(a, theta)"


class_doc_txt = """
    Foo

    Parameters
    ----------
    f : callable ``f(t, y, *f_args)``
        Aaa.
    jac : callable ``jac(t, y, *jac_args)``

        Bbb.

    Attributes
    ----------
    t : float
        Current time.
    y : ndarray
        Current variable values.

        * hello
        * world
    an_attribute : float
        The docstring is printed instead
    no_docstring : str
        But a description
    no_docstring2 : str
    multiline_sentence
    midword_period
    no_period

    Methods
    -------
    a
    b
    c

    Examples
    --------
    For usage examples, see `ode`.
"""


def test_class_members_doc():
    doc = ClassDoc(None, class_doc_txt)
    line_by_line_compare(
        str(doc),
        """
    Foo

    Parameters
    ----------
    f : callable ``f(t, y, *f_args)``
        Aaa.
    jac : callable ``jac(t, y, *jac_args)``
        Bbb.

    Examples
    --------
    For usage examples, see `ode`.

    Attributes
    ----------
    t : float
        Current time.
    y : ndarray
        Current variable values.

        * hello
        * world
    an_attribute : float
        The docstring is printed instead
    no_docstring : str
        But a description
    no_docstring2 : str
    multiline_sentence
    midword_period
    no_period

    Methods
    -------
    a
    b
    c

    """,
    )


def test_class_members_doc_sphinx():
    class Foo:
        @property
        def an_attribute(self):
            """Test attribute"""
            return None

        @property
        def no_docstring(self):
            return None

        @property
        def no_docstring2(self):
            return None

        @property
        def multiline_sentence(self):
            """This is a
            sentence. It spans multiple lines."""
            return None

        @property
        def midword_period(self):
            """The sentence for numpy.org."""
            return None

        @property
        def no_period(self):
            """This does not have a period
            so we truncate its summary to the first linebreak

            Apparently.
            """
            return None

    doc = SphinxClassDoc(Foo, class_doc_txt)
    line_by_line_compare(
        str(doc),
        """
    Foo

    :Parameters:

        **f** : callable ``f(t, y, *f_args)``
            Aaa.

        **jac** : callable ``jac(t, y, *jac_args)``
            Bbb.

    .. rubric:: Examples

    For usage examples, see `ode`.

    :Attributes:

        **t** : float
            Current time.

        **y** : ndarray
            Current variable values.

            * hello
            * world

        :obj:`an_attribute <an_attribute>` : float
            Test attribute

        **no_docstring** : str
            But a description

        **no_docstring2** : str
            ..

        :obj:`multiline_sentence <multiline_sentence>`
            This is a sentence.

        :obj:`midword_period <midword_period>`
            The sentence for numpy.org.

        :obj:`no_period <no_period>`
            This does not have a period

    .. rubric:: Methods

    =====  ==========
    **a**
    **b**
    **c**
    =====  ==========

    """,
    )


def test_class_attributes_as_member_list():
    class Foo:
        """
        Class docstring.

        Attributes
        ----------
        an_attribute
            Another description that is not used.

        """

        @property
        def an_attribute(self):
            """Test attribute"""
            return None

    attr_doc = """:Attributes:

    :obj:`an_attribute <an_attribute>`
        Test attribute"""

    assert attr_doc in str(SphinxClassDoc(Foo))
    assert "Another description" not in str(SphinxClassDoc(Foo))

    attr_doc2 = """.. rubric:: Attributes

.. autosummary::
   :toctree:

   an_attribute"""

    cfg = dict(attributes_as_param_list=False)
    assert attr_doc2 in str(SphinxClassDoc(Foo, config=cfg))
    assert "Another description" not in str(SphinxClassDoc(Foo, config=cfg))


def test_templated_sections():
    doc = SphinxClassDoc(
        None,
        class_doc_txt,
        config={"template": jinja2.Template("{{examples}}\n{{parameters}}")},
    )
    line_by_line_compare(
        str(doc),
        """
    .. rubric:: Examples

    For usage examples, see `ode`.

    :Parameters:

        **f** : callable ``f(t, y, *f_args)``
            Aaa.

        **jac** : callable ``jac(t, y, *jac_args)``
            Bbb.

    """,
    )


def test_nonstandard_property():
    # test discovery of a property that does not satisfy isinstace(.., property)

    class SpecialProperty:
        def __init__(self, axis=0, doc=""):
            self.axis = axis
            self.__doc__ = doc

        def __get__(self, obj, type):
            if obj is None:
                # Only instances have actual _data, not classes
                return self
            else:
                return obj._data.axes[self.axis]

        def __set__(self, obj, value):
            obj._set_axis(self.axis, value)

    class Dummy:
        attr = SpecialProperty(doc="test attribute")

    doc = get_doc_object(Dummy)
    assert "test attribute" in str(doc)


def test_args_and_kwargs():
    cfg = dict()
    doc = SphinxDocString(
        """
    Parameters
    ----------
    param1 : int
        First parameter
    *args : tuple
        Arguments
    **kwargs : dict
        Keyword arguments
    """,
        config=cfg,
    )
    line_by_line_compare(
        str(doc),
        r"""
:Parameters:

    **param1** : int
        First parameter

    **\*args** : tuple
        Arguments

    **\*\*kwargs** : dict
        Keyword arguments
    """,
    )


def test_autoclass():
    cfg = dict(show_class_members=True, show_inherited_class_members=True)
    doc = SphinxClassDoc(
        str,
        """
A top section before

.. autoclass:: str
    """,
        config=cfg,
    )
    line_by_line_compare(
        str(doc),
        r"""
A top section before

.. autoclass:: str

.. rubric:: Methods


    """,
        5,
    )


xref_doc_txt = """
Test xref in Parameters, Other Parameters and Returns

Parameters
----------
p1 : int
    Integer value

p2 : float, optional
    Integer value

Other Parameters
----------------
p3 : list[int]
    List of integers
p4 : :class:`pandas.DataFrame`
    A dataframe
p5 : sequence of `int`
    A sequence

Returns
-------
out : array
    Numerical return value
"""


xref_doc_txt_expected = r"""
Test xref in Parameters, Other Parameters and Returns


:Parameters:

    **p1** : :class:`python:int`
        Integer value

    **p2** : :class:`python:float`, optional
        Integer value

:Returns:

    **out** : :obj:`array <numpy.ndarray>`
        Numerical return value


:Other Parameters:

    **p3** : :class:`python:list`\[:class:`python:int`]
        List of integers

    **p4** : :class:`pandas.DataFrame`
        A dataframe

    **p5** : :obj:`python:sequence` of `int`
        A sequence
"""


def test_xref():
    xref_aliases = {
        "sequence": ":obj:`python:sequence`",
    }

    class Config:
        def __init__(self, a, b):
            self.numpydoc_xref_aliases = a
            self.numpydoc_xref_aliases_complete = b
            # numpydoc.update_config fails if this config option not present
            self.numpydoc_validation_checks = set()
            self.numpydoc_validation_exclude = set()
            self.numpydoc_validation_overrides = dict()

    xref_aliases_complete = deepcopy(DEFAULT_LINKS)
    for key in xref_aliases:
        xref_aliases_complete[key] = xref_aliases[key]
    config = Config(xref_aliases, xref_aliases_complete)
    app = namedtuple("config", "config")(config)
    update_config(app)

    xref_ignore = {"of", "default", "optional"}

    doc = SphinxDocString(
        xref_doc_txt,
        config=dict(
            xref_param_type=True,
            xref_aliases=xref_aliases_complete,
            xref_ignore=xref_ignore,
        ),
    )

    line_by_line_compare(str(doc), xref_doc_txt_expected)


def test__error_location_no_name_attr():
    """
    Ensure that NumpyDocString._error_location doesn't fail when self._obj
    does not have a __name__ attr.

    See gh-362
    """
    from collections.abc import Callable

    # Create a Callable that doesn't have a __name__ attribute
    class Foo:
        def __call__(self):
            pass

    foo = Foo()  # foo is a Callable, but no a function instance
    assert isinstance(foo, Callable)

    # Create an NumpyDocString instance to call the _error_location method
    nds = get_doc_object(foo)

    msg = "Potentially wrong underline length.*Foo.*"
    with pytest.raises(ValueError, match=msg):
        nds._error_location(msg=msg)


def test_class_docstring_cached_property():
    """Ensure that properties marked with the `cached_property` decorator
    are listed in the Methods section. See gh-432."""
    from functools import cached_property

    class Foo:
        _x = [1, 2, 3]

        @cached_property
        def val(self):
            return self._x

    class_docstring = get_doc_object(Foo)
    assert len(class_docstring["Attributes"]) == 1
    assert class_docstring["Attributes"][0].name == "val"


def test_namedtuple_no_duplicate_attributes():
    """
    Ensure that attributes of namedtuples are not duplicated in the docstring.

    See gh-257
    """
    from collections import namedtuple

    foo = namedtuple("Foo", ("bar", "baz"))

    # Create the SphinxClassDoc object via get_doc_object
    sds = get_doc_object(foo)
    assert sds["Attributes"] == []


def test_namedtuple_class_docstring():
    """Ensure that class docstring is preserved when inheriting from namedtuple.

    See gh-257
    """
    from collections import namedtuple

    foo = namedtuple("Foo", ("bar", "baz"))

    class MyFoo(foo):
        """MyFoo's class docstring"""

    # Create the SphinxClassDoc object via get_doc_object
    sds = get_doc_object(MyFoo)
    assert sds["Summary"] == ["MyFoo's class docstring"]

    # Example dataclass where constructor params are documented explicit.
    # Parameter names/descriptions should be included in the docstring, but
    # should not result in a duplicated `Attributes` section
    class MyFooWithParams(foo):
        """
        MyFoo's class docstring

        Parameters
        ----------
        bar : str
           The bar attribute
        baz : str
           The baz attribute
        """

        bar: str
        baz: str

    sds = get_doc_object(MyFooWithParams)
    assert "MyFoo's class docstring" in sds["Summary"]
    assert len(sds["Attributes"]) == 0
    assert len(sds["Parameters"]) == 2
    assert sds["Parameters"][0].desc[0] == "The bar attribute"
    assert sds["Parameters"][1].desc[0] == "The baz attribute"


if __name__ == "__main__":
    import pytest

    pytest.main()



"""Run numpydoc validation on contents of a file."""

import argparse
import ast
import configparser
import os
import re
import sys

try:
    import tomllib
except ImportError:
    import tomli as tomllib

from pathlib import Path
from typing import Sequence, Tuple, Union

from tabulate import tabulate

from .. import docscrape, validate
from .utils import find_project_root


class AstValidator(validate.Validator):
    """
    Overrides the :class:`Validator` to work entirely with the AST.

    Parameters
    ----------
    ast_node : ast.AST
        The node under inspection.
    filename : os.PathLike
        The file where the node is defined.
    obj_name : str
        A name for the node to use in the listing of issues for the file as a whole.
    """

    def __init__(
        self, *, ast_node: ast.AST, filename: os.PathLike, obj_name: str
    ) -> None:
        self.node: ast.AST = ast_node
        self.raw_doc: str = ast.get_docstring(self.node, clean=False) or ""
        self.clean_doc: str = ast.get_docstring(self.node, clean=True)
        self.doc: docscrape.NumpyDocString = docscrape.NumpyDocString(self.raw_doc)

        self._source_file: os.PathLike = Path(filename).resolve()
        self._name: str = obj_name

        self.is_class: bool = isinstance(ast_node, ast.ClassDef)
        self.is_module: bool = isinstance(ast_node, ast.Module)

    @staticmethod
    def _load_obj(name):
        raise NotImplementedError("AstValidator does not support this method.")

    @property
    def name(self) -> str:
        return self._name

    @property
    def is_function_or_method(self) -> bool:
        return isinstance(self.node, (ast.FunctionDef, ast.AsyncFunctionDef))

    @property
    def is_generator_function(self) -> bool:
        if not self.is_function_or_method:
            return False
        for child in ast.iter_child_nodes(self.node):
            if isinstance(child, ast.Expr) and isinstance(child.value, ast.Yield):
                return True
        return False

    @property
    def type(self) -> str:
        if self.is_function_or_method:
            return "function"
        if self.is_class:
            return "type"
        if self.is_module:
            return "module"
        raise ValueError("Unknown type.")

    @property
    def source_file_name(self) -> str:
        return self._source_file

    @property
    def source_file_def_line(self) -> int:
        return self.node.lineno if not self.is_module else 1

    @property
    def signature_parameters(self) -> Tuple[str]:
        def extract_signature(node):
            args_node = node.args
            params = []
            for arg_type in ["posonlyargs", "args", "vararg", "kwonlyargs", "kwarg"]:
                entries = getattr(args_node, arg_type)
                if arg_type in ["vararg", "kwarg"]:
                    if entries and arg_type == "vararg":
                        params.append(f"*{entries.arg}")
                    if entries and arg_type == "kwarg":
                        params.append(f"**{entries.arg}")
                else:
                    params.extend([arg.arg for arg in entries])
            params = tuple(params)
            if params and params[0] in {"self", "cls"}:
                return params[1:]
            return params

        params = tuple()
        if self.is_function_or_method:
            params = extract_signature(self.node)
        elif self.is_class:
            for child in self.node.body:
                if isinstance(child, ast.FunctionDef) and child.name == "__init__":
                    params = extract_signature(child)
        return params

    @property
    def method_source(self) -> str:
        with open(self.source_file_name) as file:
            source = ast.get_source_segment(file.read(), self.node)
        return source


class DocstringVisitor(ast.NodeVisitor):
    """
    Visits nodes in the AST from a given module and reporting numpydoc issues.

    Parameters
    ----------
    filepath : str
        The absolute or relative path to the file to inspect.
    config : dict
        Configuration options for reviewing flagged issues.
    """

    def __init__(
        self,
        filepath: str,
        config: dict,
    ) -> None:
        self.config: dict = config
        self.filepath: str = filepath
        self.module_name: str = Path(self.filepath).stem
        self.stack: list[str] = []
        self.findings: list = []

    def _ignore_issue(self, node: ast.AST, check: str) -> bool:
        """
        Check whether the issue should be ignored.

        Parameters
        ----------
        node : ast.AST
            The node under inspection.
        check : str
            The code for the check being evaluated.

        Return
        ------
        bool
            Whether the issue should be excluded from the report.
        """
        if check not in self.config["checks"]:
            return True

        if self.config["overrides"]:
            try:
                pattern = self.config["overrides"][check]
                if re.search(pattern, ast.get_docstring(node)) is not None:
                    return True
            except KeyError:
                pass

        return False

    def _get_numpydoc_issues(self, node: ast.AST) -> None:
        """
        Get numpydoc validation issues.

        Parameters
        ----------
        node : ast.AST
            The node under inspection.
        """
        name = ".".join(self.stack)
        report = validate.validate(
            name, AstValidator, ast_node=node, filename=self.filepath
        )
        self.findings.extend(
            [
                [f'{self.filepath}:{report["file_line"]}', name, check, description]
                for check, description in report["errors"]
                if not self._ignore_issue(node, check)
            ]
        )

    def visit(self, node: ast.AST) -> None:
        """
        Visit a node in the AST and report on numpydoc validation issues.

        Parameters
        ----------
        node : ast.AST
            The node to visit.
        """
        if isinstance(
            node, (ast.Module, ast.ClassDef, ast.FunctionDef, ast.AsyncFunctionDef)
        ):
            self.stack.append(
                self.module_name if isinstance(node, ast.Module) else node.name
            )

            if not (
                self.config["exclude"]
                and re.search(self.config["exclude"], ".".join(self.stack))
            ):
                self._get_numpydoc_issues(node)

            self.generic_visit(node)
            _ = self.stack.pop()


def parse_config(dir_path: os.PathLike = None) -> dict:
    """
    Parse config information from a pyproject.toml or setup.cfg file.

    This function looks in the provided directory path first for a
    pyproject.toml file. If it finds that, it won't look for a setup.cfg
    file.

    Parameters
    ----------
    dir_path : os.PathLike
        An absolute or relative path to a directory containing
        either a pyproject.toml file specifying a
        [tool.numpydoc_validation] section or a setup.cfg file
        specifying a [tool:numpydoc_validation] section.
        For example, ``~/my_project``. If not provided, the hook
        will try to find the project root directory.

    Returns
    -------
    dict
        Config options for the numpydoc validation hook.
    """
    options = {"checks": {"all"}, "exclude": set(), "overrides": {}}
    dir_path = Path(dir_path).expanduser().resolve()

    toml_path = dir_path / "pyproject.toml"
    cfg_path = dir_path / "setup.cfg"

    def compile_regex(expressions):
        return (
            re.compile(r"|".join(exp for exp in expressions if exp))
            if expressions
            else None
        )

    def extract_check_overrides(options, config_items):
        for option, value in config_items:
            if option.startswith("override_"):
                _, check = option.split("_")
                if value:
                    options["overrides"][check.upper()] = compile_regex(value)

    if toml_path.is_file():
        with open(toml_path, "rb") as toml_file:
            pyproject_toml = tomllib.load(toml_file)
            config = pyproject_toml.get("tool", {}).get("numpydoc_validation", {})
            options["checks"] = set(config.get("checks", options["checks"]))

            global_exclusions = config.get("exclude", options["exclude"])
            options["exclude"] = set(
                global_exclusions
                if not isinstance(global_exclusions, str)
                else [global_exclusions]
            )

            extract_check_overrides(options, config.items())

    elif cfg_path.is_file():
        config = configparser.ConfigParser()
        config.read(cfg_path)
        numpydoc_validation_config_section = "tool:numpydoc_validation"
        try:
            try:
                options["checks"] = set(
                    config.get(numpydoc_validation_config_section, "checks")
                    .rstrip(",")
                    .split(",")
                    or options["checks"]
                )
            except configparser.NoOptionError:
                pass
            try:
                options["exclude"] = set(
                    config.get(numpydoc_validation_config_section, "exclude")
                    .rstrip(",")
                    .split(",")
                    or options["exclude"]
                )
            except configparser.NoOptionError:
                pass

            extract_check_overrides(
                options, config.items(numpydoc_validation_config_section)
            )

        except configparser.NoSectionError:
            pass

    options["checks"] = validate.get_validation_checks(options["checks"])
    options["exclude"] = compile_regex(options["exclude"])
    return options


def process_file(filepath: os.PathLike, config: dict) -> "list[list[str]]":
    """
    Run numpydoc validation on a file.

    Parameters
    ----------
    filepath : path-like
        The absolute or relative path to the file to inspect.
    config : dict
        Configuration options for reviewing flagged issues.

    Returns
    -------
    list[list[str]]
        A list of [name, check, description] lists for flagged issues.
    """
    with open(filepath) as file:
        module_node = ast.parse(file.read(), filepath)

    docstring_visitor = DocstringVisitor(filepath=str(filepath), config=config)
    docstring_visitor.visit(module_node)

    return docstring_visitor.findings


def main(argv: Union[Sequence[str], None] = None) -> int:
    """Run the numpydoc validation hook."""

    project_root_from_cwd, config_file = find_project_root(["."])
    config_options = parse_config(project_root_from_cwd)
    ignored_checks = (
        "\n  "
        + "\n  ".join(
            [
                f"- {check}: {validate.ERROR_MSGS[check]}"
                for check in set(validate.ERROR_MSGS.keys()) - config_options["checks"]
            ]
        )
        + "\n"
    )

    parser = argparse.ArgumentParser(
        description="Run numpydoc validation on files with option to ignore individual checks.",
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser.add_argument(
        "files", type=str, nargs="+", help="File(s) to run numpydoc validation on."
    )
    parser.add_argument(
        "--config",
        type=str,
        help=(
            "Path to a directory containing a pyproject.toml or setup.cfg file.\n"
            "The hook will look for it in the root project directory.\n"
            "If both are present, only pyproject.toml will be used.\n"
            "Options must be placed under\n"
            "    - [tool:numpydoc_validation] for setup.cfg files and\n"
            "    - [tool.numpydoc_validation] for pyproject.toml files."
        ),
    )
    parser.add_argument(
        "--ignore",
        type=str,
        nargs="*",
        help=(
            f"""Check codes to ignore.{
                ' Currently ignoring the following from '
                f'{Path(project_root_from_cwd) / config_file}: {ignored_checks}'
                'Values provided here will be in addition to the above, unless an alternate config is provided.'
                if config_options["checks"] else ''
            }"""
        ),
    )

    args = parser.parse_args(argv)
    project_root, _ = find_project_root(args.files)
    config_options = parse_config(args.config or project_root)
    config_options["checks"] -= set(args.ignore or [])

    findings = []
    for file in args.files:
        findings.extend(process_file(file, config_options))

    if findings:
        print(
            tabulate(
                findings,
                headers=["file", "item", "check", "description"],
                tablefmt="grid",
                maxcolwidths=50,
            ),
            file=sys.stderr,
        )
        return 1
    return 0


if __name__ == "__main__":
    raise SystemExit(main())



import pytest
from collections import defaultdict
from io import StringIO
from pathlib import PosixPath
from copy import deepcopy

from docutils import nodes

from numpydoc.numpydoc import (
    mangle_docstrings,
    _clean_text_signature,
    update_config,
    clean_backrefs,
)
from numpydoc.xref import DEFAULT_LINKS
from sphinx.ext.autodoc import ALL
from sphinx.util import logging


class MockConfig:
    numpydoc_use_plots = False
    numpydoc_show_class_members = True
    numpydoc_show_inherited_class_members = True
    numpydoc_class_members_toctree = True
    numpydoc_xref_param_type = False
    numpydoc_xref_aliases = {}
    numpydoc_xref_aliases_complete = deepcopy(DEFAULT_LINKS)
    numpydoc_xref_ignore = set()
    templates_path = []
    numpydoc_citation_re = "[a-z0-9_.-]+"
    numpydoc_attributes_as_param_list = True
    numpydoc_validation_checks = set()
    numpydoc_validation_exclude = set()
    numpydoc_validation_overrides = dict()


class MockBuilder:
    config = MockConfig()


class MockApp:
    config = MockConfig()
    builder = MockBuilder()
    translator = None

    def __init__(self):
        self.builder.app = self
        # Attrs required for logging
        self.verbosity = 2
        self._warncount = 0
        self.warningiserror = False


def test_mangle_docstrings_basic():
    s = """
A top section before

.. autoclass:: str
    """
    lines = s.split("\n")
    mangle_docstrings(MockApp(), "class", "str", str, {}, lines)
    assert "rpartition" in [x.strip() for x in lines]

    lines = s.split("\n")
    mangle_docstrings(MockApp(), "class", "str", str, {"members": ["upper"]}, lines)
    assert "rpartition" not in [x.strip() for x in lines]
    assert "upper" in [x.strip() for x in lines]

    lines = s.split("\n")
    mangle_docstrings(MockApp(), "class", "str", str, {"exclude-members": ALL}, lines)
    assert "rpartition" not in [x.strip() for x in lines]
    assert "upper" not in [x.strip() for x in lines]

    lines = s.split("\n")
    mangle_docstrings(
        MockApp(), "class", "str", str, {"exclude-members": ["upper"]}, lines
    )
    assert "rpartition" in [x.strip() for x in lines]
    assert "upper" not in [x.strip() for x in lines]


def test_mangle_docstrings_inherited_class_members():
    # if subclass docs are rendered, this PosixPath should have Path.samefile
    p = """
A top section before

.. autoclass:: pathlib.PosixPath
"""
    lines = p.split("\n")
    app = MockApp()
    mangle_docstrings(app, "class", "pathlib.PosixPath", PosixPath, {}, lines)
    lines = [x.strip() for x in lines]
    assert "samefile" in lines
    app.config.numpydoc_show_inherited_class_members = False
    lines = p.split("\n")
    mangle_docstrings(app, "class", "pathlib.PosixPath", PosixPath, {}, lines)
    lines = [x.strip() for x in lines]
    assert "samefile" not in lines
    app.config.numpydoc_show_inherited_class_members = dict()
    lines = p.split("\n")
    mangle_docstrings(app, "class", "pathlib.PosixPath", PosixPath, {}, lines)
    lines = [x.strip() for x in lines]
    assert "samefile" in lines
    app.config.numpydoc_show_inherited_class_members = defaultdict(bool)
    lines = p.split("\n")
    mangle_docstrings(app, "class", "pathlib.PosixPath", PosixPath, {}, lines)
    lines = [x.strip() for x in lines]
    assert "samefile" not in lines


def test_clean_text_signature():
    assert _clean_text_signature(None) is None
    assert _clean_text_signature("func($self)") == "func()"
    assert (
        _clean_text_signature("func($self, *args, **kwargs)") == "func(*args, **kwargs)"
    )
    assert _clean_text_signature("($self)") == "()"
    assert _clean_text_signature("()") == "()"
    assert _clean_text_signature("func()") == "func()"
    assert (
        _clean_text_signature("func($self, /, *args, **kwargs)")
        == "func(*args, **kwargs)"
    )
    assert (
        _clean_text_signature("func($self, other, /, *args, **kwargs)")
        == "func(other, *args, **kwargs)"
    )
    assert _clean_text_signature("($module)") == "()"
    assert _clean_text_signature("func($type)") == "func()"
    assert (
        _clean_text_signature('func($self, foo="hello world")')
        == 'func(foo="hello world")'
    )
    assert (
        _clean_text_signature("func($self, foo='hello world')")
        == "func(foo='hello world')"
    )
    assert _clean_text_signature('func(foo="hello world")') == 'func(foo="hello world")'
    assert _clean_text_signature('func(foo="$self")') == 'func(foo="$self")'
    assert _clean_text_signature('func($self, foo="$self")') == 'func(foo="$self")'
    assert _clean_text_signature("func(self, other)") == "func(self, other)"
    assert _clean_text_signature("func($self, *args)") == "func(*args)"


@pytest.fixture
def f():
    def _function_without_seealso_and_examples():
        """
        A function whose docstring has no examples or see also section.

        Expect SA01 and EX01 errors if validation enabled.
        """
        pass

    return _function_without_seealso_and_examples


@pytest.mark.parametrize(
    (
        "numpydoc_validation_checks",
        "expected_warn",
        "non_warnings",
    ),
    (
        # Validation configured off - expect no warnings
        (set(), [], []),
        # Validation on with expected warnings
        ({"SA01", "EX01"}, ("SA01", "EX01"), []),
        # Validation on with only one activated check
        ({"SA01"}, ("SA01",), ("EX01",)),
    ),
)
def test_mangle_docstring_validation_warnings(
    f,
    numpydoc_validation_checks,
    expected_warn,
    non_warnings,
):
    app = MockApp()
    # Set up config for test
    app.config.numpydoc_validation_checks = numpydoc_validation_checks
    # Update configuration
    update_config(app)
    # Set up logging
    status, warning = StringIO(), StringIO()
    logging.setup(app, status, warning)
    # Run mangle docstrings with the above configuration
    mangle_docstrings(app, "function", "f", f, None, f.__doc__.split("\n"))
    # Assert that all (and only) expected warnings are logged
    warnings = warning.getvalue()
    for w in expected_warn:
        assert w in warnings
    for w in non_warnings:
        assert w not in warnings


def test_mangle_docstring_validation_exclude():
    def function_with_bad_docstring():
        """
        This docstring will raise docstring validation warnings."""

    app = MockApp()
    app.config.numpydoc_validation_checks = {"all"}
    app.config.numpydoc_validation_exclude = [r"_bad_"]
    # Call update_config to construct regexp from config value
    update_config(app)
    # Setup for catching warnings
    status, warning = StringIO(), StringIO()
    logging.setup(app, status, warning)
    # Run mangle docstrings on function_with_bad_docstring
    mangle_docstrings(
        app,
        "function",
        function_with_bad_docstring.__name__,
        function_with_bad_docstring,
        None,
        function_with_bad_docstring.__doc__.split("\n"),
    )
    # Validation is skipped due to exclude pattern matching fn name, therefore
    # no warnings expected
    assert warning.getvalue() == ""


@pytest.mark.parametrize("overrides", [{}, {"SS02"}, {"SS02", "SS03"}])
def test_mangle_docstrings_overrides(overrides):
    def process_something_noop_function():
        """Process something."""

    app = MockApp()
    app.config.numpydoc_validation_checks = {"all"}
    app.config.numpydoc_validation_overrides = {
        check: [r"^Process "]  # overrides are regex on docstring content
        for check in overrides
    }
    update_config(app)

    # Setup for catching warnings
    status, warning = StringIO(), StringIO()
    logging.setup(app, status, warning)

    # Run mangle docstrings on process_something_noop_function
    mangle_docstrings(
        app,
        "function",
        process_something_noop_function.__name__,
        process_something_noop_function,
        None,
        process_something_noop_function.__doc__.split("\n"),
    )

    findings = warning.getvalue()
    assert " EX01: " in findings  # should always be there
    for check in overrides:
        assert f" {check}: " not in findings


def test_update_config_invalid_validation_set():
    app = MockApp()
    # Results in {'a', 'l'} instead of {"all"}
    app.config.numpydoc_validation_checks = set("all")
    with pytest.raises(ValueError, match="Unrecognized validation code"):
        update_config(app)


def test_update_config_exclude_str():
    app = MockApp()
    app.config.numpydoc_validation_checks = set()
    app.config.numpydoc_validation_exclude = "shouldnt-be-a-str"
    with pytest.raises(ValueError, match=r"\['shouldnt-be-a-str'\]"):
        update_config(app)


def test_clean_backrefs():
    """Check ids are not cleaned from inline backrefs."""
    par = nodes.paragraph(rawsource="", text="")
    inline_ref = nodes.inline(rawsource="", text="", ids=["id1"])
    inline_ref += nodes.reference(rawsource="", text="[1]", refid="r123-1")
    citation = nodes.citation(
        rawsource="", docname="index", backrefs=["id1"], ids=["r123-1"]
    )
    citation += nodes.label("1")
    citation += nodes.paragraph(rawsource="", text="Author. Title.")
    par += inline_ref
    par += citation
    clean_backrefs(app=MockApp(), doc=par, docname="index")
    assert "id1" in citation["backrefs"]


if __name__ == "__main__":
    import pytest

    pytest.main()



"""Test module for hook."""  # numpydoc ignore=ES01,SA01


def some_function(name):
    """Welcome to some function."""
    pass


class MyClass:
    """This is MyClass."""

    def __init__(self):
        pass

    def __repr__(self):  # numpydoc ignore=GL08
        pass

    def do_something(self, *args, **kwargs):
        """
        Do something.

        Parameters
        ----------
        *args
        """
        pass

    def process(self):
        """Process stuff."""
        pass


class NewClass:
    pass



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

from datetime import date
import numpydoc

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

import os
import sys

# for example.py
sys.path.insert(0, os.path.abspath("."))
# project root
sys.path.insert(0, os.path.abspath(".."))

os.environ["MPLBACKEND"] = "Agg"  # avoid tkinter import errors on rtfd.io

# -- Project information -----------------------------------------------------

project = "numpydoc"
copyright = f"2019-{date.today().year}, numpydoc maintainers"

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.

# version = .__version__
# The full version, including alpha/beta/rc tags.
release = numpydoc.__version__
version = numpydoc.__version__

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.doctest",
    "sphinx.ext.intersphinx",
    "sphinx.ext.todo",
    "numpydoc",
    "sphinx.ext.ifconfig",
    "sphinx.ext.viewcode",
    "sphinx.ext.imgmath",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The root toctree document
# master_doc = "index"  # NOTE: will be changed to `root_doc` in sphinx 4

numpydoc_xref_param_type = True
numpydoc_xref_ignore = {"optional", "type_without_description", "BadException"}
# Run docstring validation as part of build process
numpydoc_validation_checks = {"all", "GL01", "SA04", "RT03"}

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.

html_theme = "pydata_sphinx_theme"
html_theme_options = {
    "show_prev_next": False,
    "navbar_end": ["theme-switcher", "search-field.html", "navbar-icon-links.html"],
    "icon_links": [
        {
            "name": "GitHub",
            "url": "https://github.com/numpy/numpydoc",
            "icon": "fab fa-github-square",
            "type": "fontawesome",
        },
    ],
}

html_sidebars = {
    "**": [],
}
html_context = {
    "default_mode": "light",
}

html_title = f"{project} v{version} Manual"
html_last_updated_fmt = "%b %d, %Y"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = []  # ['_static']

# Output file base name for HTML help builder.
htmlhelp_basename = "project-templatedoc"


# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    # 'papersize': 'letterpaper',
    # The font size ('10pt', '11pt' or '12pt').
    # 'pointsize': '10pt',
    # Additional stuff for the LaTeX preamble.
    # 'preamble': '',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (
        "index",
        "numpydoc.tex",
        "numpydoc Documentation",
        "Numpydoc maintainers",
        "manual",
    ),
]

# -- Intersphinx setup ----------------------------------------------------

# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
    "numpy": ("https://numpy.org/devdocs/", None),
    "sklearn": ("https://scikit-learn.org/stable/", None),
}
